# ##### BEGIN GPL LICENSE BLOCK #####
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software Foundation,
#  Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# ##### END GPL LICENSE BLOCK #####

import bpy
from bpy.props import *
import os
import random
from mathutils import *

def act(ob):
    bpy.ops.object.select_all(action='DESELECT')
    bpy.context.scene.objects.active=ob
    ob.select=True

def layer(x):
    layers=[]
    for i in range(0,20):
        if i==x-1:
            layers.append(True)
        else:
            layers.append(False)
   
    return layers

def assignmats(ob):
    sce = bpy.context.scene
    act(ob)
    if bpy.data.materials.get('fracture') is None:
        f=bpy.data.materials.new(name='fracture')
        bpy.ops.object.material_slot_add()
        ob.material_slots[-1].material=f#lets hope this is the last
    else:
        if ob.material_slots.get('fracture') is None:
            bpy.ops.object.material_slot_add()
            sce.objects.active.material_slots[-1].material = bpy.data.materials['fracture']


def setupshards(context, stype, rb_bounds, fragility):
    sce = context.scene

    tobeprocessed = []
    for ob in sce.objects:
        if ob.select and ob.type=='MESH':
            tobeprocessed.append(ob)

    for ob in tobeprocessed:
        g = ob.game

        g.physics_type = 'RIGID_BODY'
        g.use_collision_bounds = 1
        g.collision_bounds_type = rb_bounds
        g.rotation_damping = 0.9

        sizex, sizey, sizez = getsizefrommesh(ob)
        approxvolume = sizex * sizey * sizez
        g.mass = approxvolume
        

        act(ob)
        
        if g.properties.get('fracture_shard')is None:
            
            bpy.ops.object.game_property_new(type='STRING',name='fracture_shard')
            
            bpy.ops.object.game_property_new(type='FLOAT',name='fragility')
            bpy.ops.logic.sensor_add( type='COLLISION', name='fracturecollision')
            c=g.sensors['fracturecollision']
            c.use_pulse=True
            
            bpy.ops.logic.sensor_add( type='PROPERTY', name='is_static')
            s=g.sensors['is_static']
            s.property='fracture_shard'
            s.value='STATIC'
            
            bpy.ops.logic.controller_add( type='PYTHON', name='fracture' )
            f=g.controllers['fracture']
            if bpy.data.texts.get('fracture_hierarchy_logic.py') is None:
                import_text('fracture_hierarchy_logic.py')
            f.text=bpy.data.texts['fracture_hierarchy_logic.py']
            
            bpy.ops.logic.actuator_add(type='EDIT_OBJECT', name="end", object="")
            g.actuators['end'].mode='ENDOBJECT'
            
            f.link( c )
            f.link( s )
            g.actuators['end'].link(f)
            
        g.properties['fracture_shard'].value=stype
        g.properties['fragility'].value=fragility
        
        g.radius=0.01
        
        #act(ob)


def create_cutter(context, crack_type, scale, roughness):
    ncuts = 12
    if crack_type == 'FLAT' or crack_type == 'FLAT_ROUGH':
        bpy.ops.mesh.primitive_cube_add(
            view_align=False,
            enter_editmode=False,
            location=(0, 0, 0),
            rotation=(0, 0, 0),
            layers=context.scene.layers)

        for v in context.active_object.data.vertices:
            v.co[0] += 1.0
            v.co *= scale

        bpy.ops.object.editmode_toggle()
        bpy.ops.mesh.faces_shade_smooth()
        bpy.ops.uv.reset()

        if crack_type == 'FLAT_ROUGH':
            bpy.ops.mesh.subdivide(
                number_cuts=ncuts,
                fractal=0.0,
                smoothness=0)#roughness *  scale

            bpy.ops.mesh.vertices_smooth(repeat=1)

        bpy.ops.object.editmode_toggle()
    
    if crack_type == 'SPHERE' or crack_type == 'SPHERE_ROUGH':
        bpy.ops.mesh.primitive_ico_sphere_add(subdivisions=4,
            size=1,
            view_align=False,
            enter_editmode=False,
            location=(0, 0, 0),
            rotation=(0, 0, 0),
            layers=context.scene.layers)
        bpy.ops.object.editmode_toggle()
        bpy.ops.mesh.faces_shade_smooth()
        bpy.ops.uv.smart_project(angle_limit=66, island_margin=0)

        bpy.ops.object.editmode_toggle()
        for v in context.active_object.data.vertices:
            v.co[0] += 1.0
            v.co *= scale

    if crack_type == 'SPHERE_ROUGH' or crack_type =='FLAT_ROUGH':
        for v in context.scene.objects.active.data.vertices:
            v.co[0] += roughness * scale * 0.2 * (random.random() - 0.5)
            v.co[1] += roughness * scale * 0.1 * (random.random() - 0.5)
            v.co[2] += roughness * scale * 0.1 * (random.random() - 0.5)

    bpy.context.active_object.select = True
    
    # Adding fracture material
    sce = bpy.context.scene
    bpy.ops.object.material_slot_add()
    sce.objects.active.material_slots[0].material = bpy.data.materials['fracture']
    


def getsizefrommesh(ob):
    bb = ob.bound_box
    return (
        bb[5][0] - bb[0][0],
        bb[3][1] - bb[0][1],
        bb[1][2] - bb[0][2])


def getIslands(shard):
    sm = shard.data
    vgroups = []
    fgroups = []

    vgi = []
    for v in sm.vertices:
        vgi.append(-1)
    
    gindex = 0
    for i in range(len(vgi)):
        if vgi[i] == -1:
            gproc = [i]
            vgroups.append([i])
            fgroups.append([])

            while len(gproc) > 0:
                i = gproc.pop(0)
                for f in sm.faces:
                    
                    for v in f.vertices:
                        if v == i:
                            for v1 in f.vertices:
                                if vgi[v1] == -1:
                                    vgi[v1] = gindex
                                    vgroups[gindex].append(v1)
                                    gproc.append(v1)

                            fgroups[gindex].append(f.index)

            gindex += 1

    if gindex == 1:
        shards = [shard]

    else:
        shards = []
        for gi in range(0, gindex):
            bpy.ops.object.select_all(action='DESELECT')
            act(shard)
            
            bpy.ops.object.duplicate(linked=False, mode='DUMMY')
            a = bpy.context.scene.objects.active
            sm = a.data

            bpy.ops.object.editmode_toggle()
            bpy.ops.mesh.select_all(action='DESELECT')
            bpy.ops.object.editmode_toggle()

            for x in range(len(sm.vertices) - 1, -1, -1):
                if vgi[x] != gi:
                    a.data.vertices[x].select = True


            bpy.ops.object.editmode_toggle()
            bpy.ops.mesh.delete()
            bpy.ops.object.editmode_toggle()

            bpy.ops.object.origin_set(type='ORIGIN_GEOMETRY')

            shards.append(a)

        bpy.context.scene.objects.unlink(shard)

    return shards


def boolop(ob, cutter, op):
    sce = bpy.context.scene

    fault = 0
    new_shards = []

    sizex, sizey, sizez = getsizefrommesh(ob)
    gsize = sizex + sizey + sizez

    bpy.ops.object.select_all()
    act(ob)
    cutter.select = False
    
    bpy.ops.object.duplicate(linked=False, mode='DUMMY')
    bpy.ops.object.modifier_add(type='BOOLEAN')
    a = sce.objects.active
    a.modifiers[-1].object = cutter
    a.modifiers[-1].operation = op
    bpy.ops.object.modifier_apply(apply_as='DATA', modifier=bpy.context.object.modifiers[-1].name)
    nmesh=a.data
    if len(nmesh.vertices) > 0:
        
        
        
        new_shard = sce.objects.active
        bpy.ops.object.origin_set(type='ORIGIN_GEOMETRY')

        sizex, sizey, sizez = getsizefrommesh(new_shard)
        gsize2 = sizex + sizey + sizez

        if gsize2 > gsize * 1.01:               # Size check
            #print (gsize2, gsize, ob.name, cutter.name)
            fault = 1

         # This checks whether returned shards are non-manifold.
         # Problem is, if org mesh is non-manifold, it will always fail (e.g. with Suzanne).
         # And disabling it does not seem to cause any problem...
#        elif min(mesh_utils.edge_face_count(nmesh)) < 2:    # Manifold check
#            fault = 1

        if not fault:
            new_shards = getIslands(new_shard)

        else:
            sce.objects.unlink(new_shard)

    else:
        sce.objects.unlink(a)
        fault = 2

    return fault, new_shards


def splitobject(context, ob, crack_type, roughness):
    scene = context.scene

    size = getsizefrommesh(ob)
    shards = []
    scale = max(size) * 1.3

    create_cutter(context, crack_type, scale, roughness)
    cutter = context.active_object
    cutter.location = ob.location

    cutter.location[0] += random.random() * size[0] * 0.1
    cutter.location[1] += random.random() * size[1] * 0.1
    cutter.location[2] += random.random() * size[2] * 0.1
    cutter.rotation_euler = [
        random.random() * 5000.0,
        random.random() * 5000.0,
        random.random() * 5000.0]

    act(ob)
    operations = ['INTERSECT', 'DIFFERENCE']

    for op in operations:
        fault, newshards = boolop(ob, cutter, op)

        shards.extend(newshards)
        if fault > 0:
            # Delete all shards in case of fault from previous operation.
            for s in shards:
                scene.objects.unlink(s)

            scene.objects.unlink(cutter)

            return [ob]

   
    
    bpy.context.scene.objects.unlink(cutter)

    return shards

def doHierarchy(s,parent):
    bpy.ops.object.select_all(action='DESELECT')
    bpy.ops.object.add(type='EMPTY', view_align=False, enter_editmode=False, location=(0,0,0), rotation=(0,0,0), layers=bpy.context.scene.layers)
    e=bpy.context.active_object
    bpy.ops.object.game_property_new(type='STRING',name='addshard')
    bpy.context.active_object.game.properties['addshard'].value=s.name
    e.location=s.location-parent.location
    n='add_'+s.name
    e.name=n
    e.parent=parent
    e.select=True
    f=parent.game.controllers['fracture']
    bpy.ops.logic.actuator_add(type='EDIT_OBJECT', name=n, object="")    
    e.game.actuators[n].object=s  
    e.game.actuators[n].link(f)
    e.game.physics_type='NO_COLLISION'
    e.game.radius=0.0001
    e.empty_draw_size=0.05
    return e
    
def cleanmesh():
    bpy.ops.object.editmode_toggle()
    bpy.ops.mesh.select_all(action='SELECT')
    bpy.ops.mesh.dissolve_limited(angle_limit=0.001)
    bpy.ops.mesh.select_all(action='DESELECT')
    bpy.ops.object.editmode_toggle()

def fracture_basic(context, nshards, crack_type, roughness, logic, mhierarchy, stype, shard_rb_bounds, shard_fragility, hide_layer):
    tobesplit = []
    topshards=[]
    hide = []
    if logic==False:
        mhierarchy=False
    for ob in context.scene.objects:
        if ob.select:
            tobesplit.append(ob)
            assignmats(ob)
            
            topshards.append(ob)
            
            if logic:
                bpy.ops.object.setup_fracture_shards(shard_type=stype, rb_bounds=shard_rb_bounds, fragility=shard_fragility)
        
    i = 1     # I counts shards, starts with 1 - the original object
    iter = 0  # counts iterations, to prevent eternal loops in case
              # of boolean faults
    scn=bpy.context.scene
    
    maxshards = nshards * len(tobesplit)
    for topob in topshards:
        tobesplit=[topob]
        shards=[]
        while i < maxshards and len(tobesplit) > 0 and iter < maxshards * 10:
            ob = tobesplit.pop(0)
            newshards = splitobject(context, ob, crack_type, roughness)
            
            if newshards[0] != ob:#this cleanup doesn't belong here !, should be in islands thingy   
                for s in newshards: 
                    act(s)    
                    cleanmesh()  
    
            if newshards[0] != ob and not mhierarchy and ob!= topob:
                bpy.context.scene.objects.unlink(ob)
                
            if newshards[0] != ob :#############hierarchy basic
                if logic:
                    for s in newshards:
                        act(s)
                        bpy.ops.object.setup_fracture_shards(shard_type=stype, rb_bounds=shard_rb_bounds, fragility=shard_fragility)
                
                if mhierarchy:
                    for s in newshards:
                        hide.append(s)
                        e=doHierarchy(s,ob)
                        if not ob in topshards:
                            hide.append(e)
               
                
            tobesplit.extend(newshards)
    
            if len(newshards) > 1:
                shards.extend(newshards)
                i += (len(newshards) - 1)
    
    
            iter += 1
        
        if mhierarchy==False:
            for s in shards:
                if not(scn.objects.get(s.name) is None):
                    hide.append(s)
                    if logic:
                        act(s)    
                        bpy.ops.object.setup_fracture_shards(shard_type=stype, rb_bounds=shard_rb_bounds, fragility=shard_fragility)
                        e=doHierarchy(s,topob)
        for o in hide:
            
            o.layers=layer(hide_layer)


def fracture_group(context, group, hierarchy, stype, shard_rb_bounds, shard_fragility, hide_layer):
    tobesplit = []
    shards = []

    for ob in context.scene.objects:
        if (ob.select
            and (len(ob.users_group) == 0 or ob.users_group[0].name != group)):
            tobesplit.append(ob)
            assignmats(ob)
    
    cutters = bpy.data.groups[group].objects

    # @todo This can be optimized.
    # Avoid booleans on obs where bbox doesn't intersect.
        
    i = 0
    for ob in tobesplit:
        act(ob)
        
        hide=[]
        for cutter in cutters:
            fault, newshards = boolop(ob, cutter, 'INTERSECT')
            shards.extend(newshards)

            if fault == 1:
               # Delete all shards in case of fault from previous operation.
                for s in shards:
                    bpy.context.scene.objects.unlink(s)

                return
            i += 1
        act(ob)
        if hierarchy: 
            bpy.ops.object.setup_fracture_shards(shard_type=stype, rb_bounds=shard_rb_bounds, fragility=shard_fragility)
        for s in shards:
            act(s)
            if hierarchy: 
                bpy.ops.object.setup_fracture_shards(shard_type=stype, rb_bounds=shard_rb_bounds, fragility=shard_fragility)
                e=doHierarchy(s,ob)
            s.layers=layer(hide_layer)
            

class FractureSimple(bpy.types.Operator):
    '''Split object with boolean operations for simulation, uses an object'''
    bl_idname = "object.fracture_simple"
    bl_label = "Fracture Object"
    bl_options = {'REGISTER', 'UNDO'}

    exe = BoolProperty(name="Execute",
        description="If it shall actually run, for optimal performance",
        default=False)

    logic = BoolProperty(name="Logic",
        description="Logic sets up everything needed for dynamic breaking"\
                        "replacement of objects",
        default=False)
        
    multilevel_hierarchy = BoolProperty(name="Multilevel hierarchy",
        description="Hierarchy enables object to split multiple times"\
                        " depending on conditions of collision",
        default=False)

    nshards = IntProperty(name="Number of shards",
        description="Number of shards the object should be split into",
        min=2,
        default=5)

    
   
    hide_layer = IntProperty(name="Hide Layer",
        description="where to hide hierarchy objects",
        min=1,
        max=20,
        default=20)
            
    crack_type = EnumProperty(name='Crack type',
        items=(
            ('FLAT', 'Flat', 'a'),
            ('FLAT_ROUGH', 'Flat rough', 'a'),
            ('SPHERE', 'Spherical', 'a'),
            ('SPHERE_ROUGH', 'Spherical rough', 'a')),
        description='Look of the fracture surface',
        default='FLAT')

    roughness = FloatProperty(name="Roughness",
        description="Roughness of the fracture surface",
        min=0.0,
        max=1.0,
        default=0.7)
        
    fragility = FloatProperty(name="Fragility",
        min=0.0,
        max=10.0,
        default=0.5,
        description='how easy is to release shards with collisions')

    shard_type = EnumProperty(name='type',
    items=(
        ('DYNAMIC', 'Dynamic', 'a'),
        ('STATIC', 'Static', 'a'),
        ('DEACTIVATED', 'Deactivated', 'a')),
    description='How the shards will behave in beginning',
    default='DYNAMIC')
    
    rb_bounds = EnumProperty(name='bounds',
    items=(
        ('CONVEX_HULL', 'Convex Hull', 'a'),
        ('TRIANGLE_MESH', 'Triangle Mesh', 'a'),
        ('BOX', 'Box', 'a')),
    description='Bounding shape, convex recommended',
    default='CONVEX_HULL')

    def execute(self, context):
        #getIslands(context.object)
        if self.exe:
            fracture_basic(context,
                    self.nshards,
                    self.crack_type,
                    self.roughness,
                    self.logic,
                    self.multilevel_hierarchy,
                    self.shard_type,
                    self.rb_bounds,
                    self.fragility,
                    self.hide_layer)

        return {'FINISHED'}
    
    def draw(self, context):
        layout = self.layout
        layout.prop(self, "exe")
        layout.prop(self, "logic")
        layout.prop(self, "multilevel_hierarchy")
        layout.prop(self, "nshards")
        layout.prop(self, "crack_type")
        layout.prop(self, "roughness")
        layout.prop(self, "hide_layer")
        layout.prop(self, "fragility")
        layout.prop(self, "shard_type")
        layout.prop(self, "rb_bounds")


class FractureGroup(bpy.types.Operator):
    '''Split object with boolean operations for simulation, uses a group'''
    bl_idname = "object.fracture_group"
    bl_label = "Fracture Object (Group)"
    bl_options = {'REGISTER', 'UNDO'}

    exe = BoolProperty(name="Execute",
                       description="If it shall actually run, for optimal performance",
                       default=False)

    group = StringProperty(name="Group",
                           description="Specify the group used for fracturing")

    hierarchy = BoolProperty(name="Hierarchy",
        description="Hierarchy is useful for simulation of objects" \
                    " breaking in motion",
        default=False)
        
    hide_layer = IntProperty(name="Hide Layer",
        description="where to hide hierarchy objects",
        min=1,
        max=20,
        default=20)

    fragility = FloatProperty(name="Fragility",
        min=0.0,
        max=10.0,
        default=0.5,
        description='how easy is to release shards with collisions')

    shard_type = EnumProperty(name='type',
        items=(
            ('DYNAMIC', 'Dynamic', 'a'),
            ('STATIC', 'Static', 'a'),
            ('DEACTIVATED', 'Deactivated', 'a')),
        description='How the shards will behave in beginning',
        default='DYNAMIC')
    
    rb_bounds = EnumProperty(name='bounds',
        items=(
            ('CONVEX_HULL', 'Convex Hull', 'a'),
            ('TRIANGLE_MESH', 'Triangle Mesh', 'a'),
            ('BOX', 'Box', 'a')),
        description='Bounding shape, convex recommended',
        default='CONVEX_HULL')




    def execute(self, context):
        #getIslands(context.object)

        if self.exe and self.group:
            fracture_group(context, 
                            self.group,
                            self.hierarchy,
                            self.shard_type,
                            self.rb_bounds,
                            self.fragility,
                            self.hide_layer)

        return {'FINISHED'}

    def draw(self, context):
        layout = self.layout
        layout.prop(self, "exe")
        layout.prop_search(self, "group", bpy.data, "groups")
        layout.prop(self, "hierarchy")
        layout.prop(self, "hide_layer")
        layout.prop(self, "fragility")
        layout.prop(self, "shard_type")
        layout.prop(self, "rb_bounds")
#####################################################################
# Import Functions
def import_text(textname):
    opath = "//data.blend\\Text\\" + textname
    s = os.sep
    dpath = bpy.utils.script_paths()[0] + \
        '%saddons%sobject_fracture%sdata.blend\\Text\\' % (s, s, s)

    bpy.ops.wm.link_append(
            filepath=opath,
            filename=textname,
            directory=dpath,
            filemode=1,
            link=False,
            autoselect=True,
            active_layer=True,
            instance_groups=True,
            relative_path=True)
    
def import_object(obname):
    opath = "//data.blend\\Object\\" + obname
    s = os.sep
    dpath = bpy.utils.script_paths()[0] + \
        '%saddons%sobject_fracture%sdata.blend\\Object\\' % (s, s, s)

    bpy.ops.wm.link_append(
            filepath=opath,
            filename=obname,
            directory=dpath,
            filemode=1,
            link=False,
            autoselect=True,
            active_layer=True,
            instance_groups=True,
            relative_path=True)

    for ob in bpy.context.selected_objects:
        ob.location = bpy.context.scene.cursor_location


class ImportFractureRecorder(bpy.types.Operator):
    '''Imports a rigidbody recorder'''
    bl_idname = "object.import_fracture_recorder"
    bl_label = "Add Rigidbody Recorder (Fracture)"
    bl_options = {'REGISTER', 'UNDO'}

    def execute(self, context):
        import_object("RECORDER")

        return {'FINISHED'}


class ImportFractureBomb(bpy.types.Operator):
    '''Import a bomb'''
    bl_idname = "object.import_fracture_bomb"
    bl_label = "Add Bomb (Fracture)"
    bl_options = {'REGISTER', 'UNDO'}

    def execute(self, context):
        import_object("BOMB")

        return {'FINISHED'}


class ImportFractureProjectile(bpy.types.Operator, ):
    '''Imports a projectile'''
    bl_idname = "object.import_fracture_projectile"
    bl_label = "Add Projectile (Fracture)"
    bl_options = {'REGISTER', 'UNDO'}

    def execute(self, context):
        import_object("PROJECTILE")

        return {'FINISHED'}


class SetupFractureShards(bpy.types.Operator):
    ''''''
    bl_idname = "object.setup_fracture_shards"
    bl_label = "Setup Fracture Shards"
    bl_options = {'REGISTER', 'UNDO'}
    
    fragility = FloatProperty(name="Fragility",
        min=0.0,
        max=10.0,
        default=0.5,
        description='how easy is to release shards with collisions')

    shard_type = EnumProperty(name='type',
    items=(
        ('DYNAMIC', 'Dynamic', 'a'),
        ('STATIC', 'Static', 'a'),
        ('DEACTIVATED', 'Deactivated', 'a')),
    description='How the shards will behave in beginning',
    default='DYNAMIC')
    
    rb_bounds = EnumProperty(name='bounds',
    items=(
        ('CONVEX_HULL', 'Convex Hull', 'a'),
        ('TRIANGLE_MESH', 'Triangle Mesh', 'a'),
        ('BOX', 'Box', 'a')),
    description='Bounding shape, convex recommended',
    default='CONVEX_HULL')

    #def poll(self, context):

    def execute(self, context):
        setupshards(context,
                    self.shard_type,
                    self.rb_bounds,
                    self.fragility)
        return {'FINISHED'}
            